/*
 * NAME
 *   g_vesicle_radius
 *     Calculate average radius of a lipid vesicle.
 * SYNOPSIS
 *   g_vesicle_radius -s topol.tpr -f traj.xtc -o radius.xvg
 * AUTHOR
 *   Martti Louhivuori (m.j.louhivuori@rug.nl)
 * 
 * ---
 * Copyright (C) 2008  Martti Louhivuori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <ctype.h>

#include "macros.h"
#include "smalloc.h"
#include "typedefs.h"
#include "names.h"
#include "copyrite.h"
#include "statutil.h"
#include "tpxio.h"
#include "string2.h"
#include "strdb.h"
#include "vec.h"
#include "macros.h"
#include "index.h"
#include "pbc.h"
#include "xvgr.h"
#include "futil.h"
#include "matio.h"

static const char *authors[] = {
	"Written by Martti Louhivuori (m.j.louhivuori@rug.nl)",
	"Copyright (c) 2008  Martti Louhivuori"
};
static const char *gpl[] = {
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License",
	"(version 2) as published by the Free Software Foundation."
};
void sp_print(FILE *out, const char *txt)
{
	int i, s;
	s = (int) (80 - strlen(txt)) / 2.0;
	for (i=0; i < s; i++) 
		fprintf(out, " ");
	fprintf(out, "%s\n", txt);
}


void calculate_com (rvec *com, int n, atom_id index[], rvec x[])
{
	int i;
	
	/* calculate the center of mass */
	/*   TODO: actually use masses! */
	(*com)[XX] = 0.0;
	(*com)[YY] = 0.0;
	(*com)[ZZ] = 0.0;
	for (i=0; i < n; i++) {
		rvec_add(*com, x[index[i]], *com);
	}
	svmul(1.0 / n, *com, *com);
}

real calculate_avg_radius (int n, atom_id index[], rvec x[], rvec com)
{
	rvec v;
	real r=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		r = r + norm(v);
	}
	r = r / n;
	return r;
}

real calculate_avg_deviation (real r, int n, atom_id index[], rvec x[], 
		rvec com)
{
	rvec v;
	real e=0.0, d=0.0;
	int i;

	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		d = norm(v) - r;
		e = e + d*d;
	}
	e = sqrt(e / n);
	return e;
}

int get_face(int n, rvec faces[], rvec v)
{
	real d, best=0.0;
	int i, face=-1;
	rvec u;

	unitv(v,u);
	for (i=0; i<n; i++) {
		d = distance2(faces[i], u);
		if (d < best || i == 0) {
			best = d;
			face = i;
		}
	}
	return face;
}

real *local_avg_radius (int n, atom_id index[], rvec x[], rvec com, int f, 
		rvec faces[])
{
	rvec v;
	real *r;
	int *count;
	int i, face=0;

	snew(r, f);
	snew(count, f);
	/* calculate the average distance to CoM for each face*/
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		face = get_face(f, faces, v);
		r[face] += norm(v);
		count[face] += 1;
	}
	for (i=0; i < f; i++) {
		if (count[i]) 
			r[i] /= count[i];
		else
			r[i] = -1.0;
	}
	return r;
}

real *local_avg_deviation (real r[], int n, atom_id index[], rvec x[], 
		rvec com, int f, rvec faces[])
{
	rvec v;
	real *e, d=0.0;
	int *count;
	int i, face;

	snew(e, f);
	snew(count, f);
	/* calculate the average distance to CoM */
	for (i=0; i < n; i++) {
		rvec_sub(x[index[i]], com, v);
		face = get_face(f, faces, v);
		d = norm(v) - r[face];
		e[face] += d*d;
		count[face] += 1;
	}
	for (i=0; i < f; i++) {
		if (count[i])
			e[i] = sqrt(e[i] / count[i]);
		else
			e[i] = -1.0;
	}
	return e;
}


int read_xyz (char *filename, rvec **coords)
{
	FILE *fp;
	int i, count=0;
	char curr, prev;
	real x,y,z;
	char line[256];
	rvec *foo;

	// no file to open?
	if (filename == NULL)
		return -1;
	// try to open file
	fp = fopen(filename, "r");
	if (fp == NULL) {
		fprintf(stderr, "ERROR: unable to read XYZ file '%s'\n", filename);
		exit(1);
	}
	// count lines
	while (fgets(line, 256, fp) != NULL) {
		count++;
	}
	fseek(fp, 0, 0);
	// allocate memory
	snew(foo, count + 1);
	// read co-ordinates
	i = 0;
	while (!feof(fp)) {
		fscanf(fp, "%f %f %f", &(foo[i][XX]), &(foo[i][YY]), &(foo[i][ZZ]));
		i++;
	}
	fclose(fp);
	// assign co-ordinate pointer
	*coords = foo;
	// return # of lines
	return count;
}

int gmx_vesicle_radius (int argc, char *argv[])
{
	static char *desc[] = {
		"Calculate the average radius of a lipid vesicle.[PAR]",
		"For each frame calculates the average distance from the ", 
		"selected atoms to their center, i.e. the average radius of the ",
		"vesicle.",
	};
	bool        bFirst=TRUE, bDivide=FALSE, bSeparate=FALSE;
	t_topology  top;
	t_atoms    *atoms;
	matrix      box;
	char        buf[255];
	rvec       *x, com, *xyz;
	int         ePBC, status, natom, count=0;
	int         size_vesicle;
	atom_id    *index_vesicle;
	char       *grpname_vesicle;
	real        t, t_o=0.0, lasttime;
	real        r, r_o=0.0, r_avg=0.0, diff, e, *r_local, *e_local, el, d;
	FILE       *fp;
	int         n, i, ecount=0;
	
	char *keys[] = {"Radius", "Std.dev."};
	char *keys3[] = {"Radius", "Std.dev.", "Std.dev. of radii"};
	
	t_pargs pa[] = {
		{ "-separate", TRUE, etBOOL, {&bSeparate}, 
			"report average radius and deviation for each region as well" }
	};
	t_filenm fnm[] = {
		{ efTRX, "-f", NULL, ffREAD },
		{ efTPS, "-s", NULL, ffREAD },
		{ efNDX, NULL, NULL, ffOPTRD },
		{ efXVG, NULL, "radius", ffWRITE },
		{ efXYZ, NULL, NULL, ffOPTRD },
	};
#define NFILE asize(fnm)
	CopyRight(stderr, argv[0]);
	for (i=0; i < (int)asize(authors); i++)
		sp_print(stderr, authors[i]);
	fprintf(stderr, "\n");
	for (i=0; i < (int)asize(gpl); i++)
		sp_print(stderr, gpl[i]);
	fprintf(stderr, "\n");
	parse_common_args(&argc, argv, PCA_CAN_VIEW | PCA_CAN_TIME | PCA_BE_NICE, 
			NFILE, fnm, asize(pa), pa, asize(desc), desc, 0, NULL);
	
	/* read topology */
	read_tps_conf(ftp2fn(efTPS, NFILE, fnm), buf, &top, &ePBC, &x, NULL, 
			box, FALSE);
	atoms = &(top.atoms);
	/* read vesicle group */
	printf("Select group for the vesicle\n");
	get_index(atoms, ftp2fn_null(efNDX, NFILE, fnm), 1, &size_vesicle, 
			&index_vesicle, &grpname_vesicle);

	n = read_xyz(ftp2fn_null(efXYZ, NFILE, fnm), &xyz);
	if (n >= 0) 
		bDivide = TRUE;
	
	/* write header */
	fp = xvgropen(ftp2fn(efXVG, NFILE, fnm), 
			"Avg. Radius", "Time (ps)", "nm");
	if (bDivide) {
		xvgr_legend(fp, 3, keys3);
	} else {
		xvgr_legend(fp, 2, keys);
	}

	/* main loop */
	natom = read_first_x(&status, ftp2fn(efTRX, NFILE, fnm), &t, &x, box);
	do {
		calculate_com(&com, size_vesicle, index_vesicle, x);
		r = calculate_avg_radius(size_vesicle, index_vesicle, x, com);
		e = calculate_avg_deviation(r, size_vesicle, index_vesicle, x, com);
		if (bFirst) {
			r_o = r;
			t_o = t;
			bFirst = FALSE;
		}
		r_avg += r;
		if (bDivide) {
			r_local = local_avg_radius(size_vesicle, index_vesicle, x, com, 
					n, xyz);
			el = 0.0;
			ecount = 0;
			for (i=0; i < n; i++) {
				if (r_local[i] >= 0.0) {
					d = (r_local[i] - r);
					el += d*d;
					ecount++;
				}
			}
			el = sqrt(el / ecount);
			if (bSeparate) {
				e_local = local_avg_deviation(r_local, size_vesicle, 
						index_vesicle, x, com, n, xyz);
				fprintf(fp,"%g  %g  %g  %g", t, r, e, el);
				for (i=0; i < n; i++) {
					fprintf(fp, "  %g  %g", r_local[i], e_local[i]);
				}
				fprintf(fp, "\n");
			} else {
				fprintf(fp,"%g  %g  %g  %g\n", t, r, e, el);
			}
		} else {
			fprintf(fp,"%g  %g  %g\n", t, r, e);
		}
		lasttime = t;
		count += 1;
	} while (read_next_x(status, &t, natom, x, box));

	fprintf(stderr, "\n");
	r_avg /= (float) count;
	fprintf(stdout, "Average radius: %g (over %g ps)\n", r_avg, 
			(lasttime - t_o));
	diff = abs(r - r_o) / r_o;
	if (diff > 0.1) 
		fprintf(stderr, "WARNING: radius drifted by more than 10 percent\n");

	/* close down */
	fclose(fp);
	close_trj(status);
	thanx(stderr);
	return 0;
}

